#include "paramtask.h"

#define RXLINELEN 1
#define ALLOCMEM 1
#define NUMBUFFERS (10)

extern _PTR memalign(size_t, size_t) __attribute__((C,spontaneous));

#define DMA_ABLE_BUFFER(_x)  (  (((uint32_t)(_x)) > 0x5c00000) &&  (((uint32_t)(_x)) < 0x5c040000) &&  ((((uint32_t)(_x))& 0x1f)== 0) )

typedef enum {
  originSendData = 0,
  originSendDataAlloc,
} sendOrigin_t;

typedef struct{
  uint8_t *pBuf;
  uint32_t numBytes;
  sendOrigin_t origin;
} bufferInfo_t;
  
typedef struct{
  char buf[RXLINELEN] __attribute ((aligned(32)));
  bool inuse;
} buffer_t __attribute__((packed));


//this implementation relies on having a queue of pointers that represent the current buffer
//to be sent out

queue_t outgoingQueue;
buffer_t gRxBuffers[NUMBUFFERS];
bool gTxActive = FALSE;

void transmitDone(uint32_t arg);
DEFINE_PARAMTASK(transmitDone); 

void receiveDone(uint32_t arg);
DEFINE_PARAMTASK(receiveDone); 


uint8_t *getNextRxBuffer(){
  int i;
  uint8_t *ret = NULL;
  atomic{
    for(i=0;i<NUMBUFFERS; i++){
      if(gRxBuffers[i].inuse == FALSE){
	gRxBuffers[i].inuse = TRUE;
	ret = (uint8_t *)gRxBuffers[i].buf;
	break;
      }
    }
  }
  return ret;
}

void returnRxBuffer(char *buf){
  int i;
  atomic{
    for(i=0;i<NUMBUFFERS; i++){
      if(gRxBuffers[i].buf == buf){
	gRxBuffers[i].inuse = FALSE;
      }
    }
  }
}

void initRxBuffer(){
  int i;
  atomic{
    for(i=0;i<NUMBUFFERS; i++){
      gRxBuffers[i].inuse = FALSE;
    }
  }
}

  
/*
 * Start of StdControl interface
 */

command result_t StdControl.init() {
  initqueue(&outgoingQueue, defaultQueueSize);
    
  gTxActive = FALSE;
    
  return SUCCESS;
}

command result_t StdControl.start() {
  
  uint8_t *rxBuffer = memalign(32, ((RXLINELEN+32)>>5)<<5);
  
  call BulkTxRx.BulkReceive(rxBuffer, RXLINELEN);
  return SUCCESS;
}

command result_t StdControl.stop() {
  return SUCCESS;
}

/*
 * End of StdControl interface
 */

/*
 * Start of SendDataAlloc interface
 */

/**
   * Function to allocate a buffer that is compatible with this interface.
   *
   * @return a non-NULL pointer to a buffer that is compatible with this interface if there is memory available to allocate.  Is no memory is available, NULL will be returned

   */
command uint8_t *SendDataAlloc.alloc(size_t numBytes){
  return memalign(32, ((numBytes+32)>>5)<<5);
}

command void SendDataAlloc.free(uint8_t *ptr){
  if(DMA_ABLE_BUFFER(ptr)){
    free(ptr);
  }
  else{
    //debugging...this could not have been something that we allocated...
    trace(DBG_USR1,"ERROR:  BufferedUART.SendDataAlloc.free passed errant buffer %#x\r\n",(uint32_t)ptr);
    free(ptr);
  }
}
command result_t SendDataAlloc.send(uint8_t* data, uint32_t numBytes){
  int ret;
  bufferInfo_t *pBI;
    
  if(!DMA_ABLE_BUFFER(data)){
    trace(DBG_USR1,"ERROR:  BufferedUART.SendDataAlloc.send passed nonDMAable buffer %#x\r\n",(uint32_t)data);
    return FAIL;
  }
  
  pBI = malloc(sizeof(*pBI));
  if(pBI == NULL){
    trace(DBG_USR1,"FATAL ERROR:  SendDataAlloc Unable to allocate memory for temp structure\r\n");
    return FAIL;
  }
    
  pBI->pBuf = data;
  pBI->numBytes = numBytes;
  pBI->origin = originSendDataAlloc;

  atomic{
    ret = pushqueue(&outgoingQueue, (uint32_t)pBI);// see if there's enough room for this buffer
    if(gTxActive == FALSE){
      call BulkTxRx.BulkTransmit(pBI->pBuf, pBI->numBytes);
      gTxActive = TRUE;
    }
  }
  if(!ret){
    //no room in the queue!
    free(pBI);
    return FAIL;
  }

  return SUCCESS;
}

default event result_t SendDataAlloc.sendDone(uint8_t* data, uint32_t numBytes, result_t success){

  return success;
}

/*
 *End of SendDataAlloc interface
 */

/*
 * Start of SendData interface
 */

command result_t SendData.send(uint8_t* data, uint32_t length) {
  int ret;
  bufferInfo_t *pBI;
    
  pBI = malloc(sizeof(*pBI));
  if(pBI == NULL){
    trace(DBG_USR1,"FATAL ERROR:  SendData Unable to allocate memory for temp structure\r\n");
    return FAIL;
  }
    
#if ALLOCMEM
  pBI->pBuf = memalign(32, length);
  if(pBI->pBuf){
    memcpy(pBI->pBuf,data,length);
  }
  else{
    trace(DBG_USR1,"FATAL ERROR:  Unable to allocate data memory\r\n");
    free(pBI);
    return FAIL;
  }
    
#else
  pBI->pBuf = data;
#endif
    
  pBI->numBytes = length;
  pBI->origin = originSendData;

  atomic{
    ret = pushqueue(&outgoingQueue, (uint32_t)pBI);// see if there's enough room for this buffer
    if(gTxActive == FALSE){
      call BulkTxRx.BulkTransmit(pBI->pBuf, length);
      gTxActive = TRUE;
    }
  }
  if(!ret){
    //no room in the queue!
#if ALLOCMEM 
    free(pBI->pBuf);
#endif
    free(pBI);
    return FAIL;
  }

  return SUCCESS;
}

default event result_t SendData.sendDone(uint8_t* data, uint32_t numBytes, result_t suc) {
  return suc;
}

void transmitDone(uint32_t arg){
  bufferInfo_t *pBI = (bufferInfo_t *)arg;
  switch(pBI->origin){
  case originSendDataAlloc:
    signal SendDataAlloc.sendDone(pBI->pBuf, pBI->numBytes, SUCCESS);
    break;
  case originSendData:
    signal SendData.sendDone(pBI->pBuf, pBI->numBytes, SUCCESS);
#if ALLOCMEM    
    free(pBI->pBuf);
#endif
    break;
  default:
    trace(DBG_USR1,"FATAL ERROR:  BuffereUart.c found unknown interface origin %d\r\n",pBI->origin);
    
  }

  free(pBI);
  atomic{
    if(peekqueue(&outgoingQueue, (uint32_t *)(&pBI))){
      //we still have buffers in the queue
      call BulkTxRx.BulkTransmit(pBI->pBuf, pBI->numBytes);
    }
    else{
      gTxActive = FALSE;
    }
  }
}


void receiveDone(uint32_t arg){
  bufferInfo_t *pBI = (bufferInfo_t *)arg;
  if(pBI == NULL){
    return;
  }

  signal ReceiveData.receive(pBI->pBuf, pBI->numBytes);
  returnRxBuffer(pBI->pBuf);
  free(pBI);
}
  
/*
 * Start of ReceiveData interface
 */
  
default event result_t ReceiveData.receive(uint8_t* Data, uint32_t Length) {
  return SUCCESS;
}
  
/*
 * End of ReceiveData interface
 */
  
  

/*
 * End of SendData interface
 */

async event uint8_t *BulkTxRx.BulkReceiveDone(uint8_t *RxBuffer, 
					      uint16_t NumBytes){
  
  bufferInfo_t *pBI;
  uint8_t *newBuffer;

  pBI = malloc(sizeof(*pBI));
  if(pBI == NULL){
    trace(DBG_USR1,"FATAL ERROR:  Unable to allocate memory for temp structure\r\n");
    return NULL;
  }
  pBI->pBuf = RxBuffer;
  pBI->numBytes = NumBytes;
  
  POST_PARAMTASK(receiveDone,pBI);
  
  newBuffer = getNextRxBuffer();
  if(newBuffer == NULL){
    trace(DBG_USR1,"FATAL ERROR:  Unable to allocate memory for receiveBuffer\r\n");
  }
  return newBuffer;
}
  
async event BulkTxRxBuffer_t *BulkTxRx.BulkTxRxDone(BulkTxRxBuffer_t *TxRxBuffer, uint16_t NumBytes){
  return NULL;
}
  
async event uint8_t *BulkTxRx.BulkTransmitDone(uint8_t *TxBuffer, uint16_t NumBytes){
    
  bufferInfo_t *pBI;
  if(popqueue(&outgoingQueue, (uint32_t *)(&pBI))){
    //got a buffer out of the queue...make sure that it's the same as what we just got back
    if((pBI->pBuf == TxBuffer) && (pBI->numBytes == NumBytes)){
      //got the right buffer back!!
      POST_PARAMTASK(transmitDone,pBI);
    }
    else{
      trace(DBG_USR1,"FATAL ERROR:  found unexpected buffer in queue\r\n");
    }
  }
  else{
    trace(DBG_USR1,"FATAL ERROR:  unexpected tranmit queue empty during\r\n");
  }
  return NULL;
}


  
