// $Id: ScriptAssembler.java,v 1.7 2004/10/07 02:40:06 scipio Exp $

/*									tab:4
 * "Copyright (c) 2000-2003 The Regents of the University  of California.  
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Copyright (c) 2002-2003 Intel Corporation
 * All rights reserved.
 *
 * This file is distributed under the terms in the attached INTEL-LICENSE     
 * file. If you do not find these files, copies can be found by writing to
 * Intel Research Berkeley, 2150 Shattuck Avenue, Suite 1300, Berkeley, CA, 
 * 94704.  Attention:  Intel License Inquiry.
 */
/* Authors:	Phil Levis <pal@cs.berkeley.edu>
 * Date:        Aug 8 2002
 * Desc:        Assembler for Bombilla VM.
 *
 */

/**
 * @author Phil Levis <pal@cs.berkeley.edu>
 */


package net.tinyos.script;

import java.io.*;
import java.lang.reflect.*;
import java.util.*;
import java.util.regex.*;

import net.tinyos.message.*;
import net.tinyos.packet.*;
import net.tinyos.util.*;

public class ScriptAssembler implements MessageListener {

  private Configuration config;
  private ConstantMapper mapper;

  // The pattern is:
  //   group 1: 0 or 1 labels, which are <word>:
  //   group 2: the instruction proper
  //   group 3: 0 or 1 embedded opreands, which are <number>
  Pattern regexp = Pattern.compile("(\\w+:\\s*)?(\\w+)?(\\s*\\w+)?");
  Pattern opcodePattern = Pattern.compile("(\\d*)(\\D+)(\\d*)"); 
  //Pattern regexp = Pattern.compile("(\\w+)\\s*(\\w+)?"); 
    
  public ScriptAssembler(Configuration config) {
    this.config = config;
    mapper = config.getOpcodeMap();
  }
    
  public String toHexString(AssemblyTokenizer tokenizer) throws IOException, InvalidInstructionException {
    byte[] program = toByteCodes(tokenizer);
    String bytes = "";
    for (int i = 0; i < program.length; i++) {
      String val = Integer.toHexString(program[i] & 0xff);
      if (val.length() == 1) {
	val = "0" + val;
      }
      bytes += val;
      bytes += " ";
    }
    return bytes;
  }

  public boolean hasEmbeddedOperand(String instr) {
    char ch = instr.charAt(instr.length() - 1);
    return (ch <= '9' && ch >= '0');
  }
  
  private Hashtable computeLabels(Vector program) throws InvalidInstructionException {
    int count = 0;
    Hashtable table = new Hashtable();
	
    for (int i = 0; i < program.size(); i++) {
      int width = 1; 
      String line = (String)program.elementAt(i);
      Matcher match = regexp.matcher(line);
      if (!match.matches()) {
	throw new InvalidInstructionException("Could not parse " + line);
      }
      String label = match.group(1);
      if (label == null) {label = "";}
      String opcode = match.group(2);
      String operand = match.group(3);
      //System.out.println(line + " -> " + label + " " + opcode + " " + operand);
      if (!label.equals("")) {
	label = label.trim();
	label = label.replace(':', ' ');
	label = label.trim();
	table.put(label, "" + count);
	//System.out.println("Binding " + label + " to " + count);
      }
      if (opcode != null &&
	  !opcode.equals("")) {
	char ch = opcode.charAt(0);
	if (ch >= '0' && ch <= '9') {
	  width = (ch - '0'); 
	}
	count += width;
      }
    }
    return table;
  }

  private Vector processLabels(Vector program, Hashtable labels) throws InvalidInstructionException {
    Vector processedProgram = new Vector();
    for (int i = 0; i < program.size(); i++) {
      String line = (String)program.elementAt(i);
      Matcher match = regexp.matcher(line);
      if (!match.matches()) {
	throw new InvalidInstructionException("Could not parse " + line);
      }
      String label = match.group(1);
      String opcode = match.group(2);
      String operand = match.group(3);
      if (operand == null) {operand = "";}
      else {
	//System.out.println("Operand: " + operand);
      }
      operand = operand.trim();
	
      if (labels.containsKey(operand)) {
	//System.out.println("Mapping " + operand + " to " + labels.get(operand));
	operand = (String)labels.get(operand);
      }
	
      String processed = opcode + " " + operand;
      processed = processed.trim();
      if (opcode != null &&
	  !opcode.equals("")) {
	processedProgram.add(processed); // Don't include empty lines
      }
    }
    return processedProgram;
  }
  
  private void addByteCodes(Vector dest, String instruction) throws InvalidInstructionException {

    Matcher instrMatch = regexp.matcher(instruction);
    if (!instrMatch.matches()) {
      throw new InvalidInstructionException("Could not parse " + instruction);
    }
    String opcode = instrMatch.group(2);
    String operand = instrMatch.group(3);
    //System.out.println("<" + instruction + ">  ->  " + instrMatch.group(1) + " " + opcode + " " + operand);

    if (!mapper.hasName(opcode.toUpperCase())) {
      throw new InvalidInstructionException("Instruction " + opcode + " not part of the ISA. Are you scripting in the right language?");
    }

    Matcher opMatch = opcodePattern.matcher(opcode);
    if (!opMatch.matches()) {
      throw new InvalidInstructionException("Could not parse " + opcode);
    }
		
    int width = 1;
    int embedded = 0;    
    byte opcodeVal = mapper.nameToCode(opcode.toUpperCase());
    long operandVal = 0;
    
    if (!opMatch.group(1).equals("")) {
      width = Integer.parseInt(opMatch.group(1));
    }
    if (!opMatch.group(3).equals("")) {
      embedded = Integer.parseInt(opMatch.group(3));
    }

    //System.out.println(opcode + " has width " + width);
    if (operand != null) {
      operand = operand.trim();
      operandVal = Long.parseLong(operand);
      //System.out.println("Operand for " + opcode + " is " + operandVal);
    }

	
    if (operandVal > (1 << embedded)) {
      throw new InvalidInstructionException(opcode + " has " + width + " bits of operand, but " + operand + " was specified (too large).");
    }

    byte[] result = new byte[width];
    result[0] = opcodeVal;
    
    for (int i = width - 1; i >=0 ; i--) {
      //System.out.print("Adding operand: " + result[i] + " -> ");
      result[i] += (byte)(operandVal & 0xff);
      //System.out.println("" + result[i]);
      operandVal = operandVal >> 8;
    }

    for (int i = 0; i < result.length; i++) {
      dest.addElement(new Byte(result[i]));
    }
  }
    
  public byte[] toByteCodes(AssemblyTokenizer tokenizer) throws IOException, InvalidInstructionException {
    Vector program = new Vector();
    //System.out.println("Adding instructions");
    while (tokenizer.hasMoreInstructions()) {
      String instr = tokenizer.nextLine();
      if (!instr.equals("")) {
	program.add(instr);
      }
    }
    //System.out.println("Computing labels.");
    Hashtable labels = computeLabels(program);
    //System.out.println("Processing labels.");
    Vector program2 = processLabels(program, labels);
    Vector bytecodes = new Vector();

    for (int i = 0; i < program2.size(); i++) {
      String instr = (String)program2.elementAt(i);
      addByteCodes(bytecodes, instr);
    }
      
    byte result[] = new byte[bytecodes.size()];
    for (int i = 0; i < bytecodes.size(); i++) {
      Byte code = (Byte)bytecodes.elementAt(i);
      result[i] = code.byteValue();

    }
    //System.err.println();
    return result;
  }

  public static void usage() {
    System.err.println("usage: ScriptAssembler [-h|-comm <source>] <context> <version> <file>");
  }

  public void messageReceived(int to, Message message) {
    BombillaErrorMsg msg = (BombillaErrorMsg)message;
    String context = config.getCapsuleMap().codeToName((byte)msg.get_context());
    String cause = config.getErrorMap().codeToName((byte)msg.get_reason());
    String capsule = config.getCapsuleMap().codeToName((byte)msg.get_capsule());
    String instruction = "" + msg.get_instruction();
    
    System.err.println("Error received:");
    System.err.println("  Context:     " + context);
    System.err.println("  Cause:       " + cause);
    System.err.println("  Capsule:     " + capsule);
    System.err.println("  Instruction: " + instruction);
    System.err.println();
  }
  
  public static void main(String[] args) {
    String asmFile = null;
    String source = null;
    String contextName = null;
    String versionString = null;
    
    for (int i = 0; i < args.length; i++) {
      String arg = args[i];
      if (arg.equals("-h") || arg.equals("--help")) {
	usage();
	System.exit(0);
      }
      else if (arg.equals("-comm")) {
	if (i == (args.length - 1)) {
	  usage();
	  System.exit(1);
	}
	i++;
	source = args[i];
      }
      else if (contextName == null) {
	if (i >= (args.length - 2)) { // Need more arguments
	  usage();
	  System.exit(1);
	}
	contextName = args[i];
	i++;
	versionString = args[i];
	i++;
	asmFile = args[i];
      }
      else {
	usage();
	System.exit(1);
      }
    }
    if (contextName == null || asmFile == null) {
      usage();
      System.exit(1);
    }
    try {
      /* Compile the code to bytecodes */
      Configuration config;
      ScriptAssembler assembler;
      AssemblyTokenizer tokenizer;
      byte[] code;	
      config = new Configuration("vm.vmdf");
      assembler = new ScriptAssembler(config);
      tokenizer = new AssemblyTokenizer(new FileReader(asmFile));
      code = assembler.toByteCodes(tokenizer);

      /* Build the object to inject the code, listen for errors. */
      PhoenixSource phoenix;
      MoteIF moteIF;
      ScriptInjector injector;
      if (source != null) {
	phoenix = BuildSource.makePhoenix(PrintStreamMessenger.err);
      }
      else {
	phoenix = BuildSource.makePhoenix(source, PrintStreamMessenger.err);
      }
      moteIF = new MoteIF(phoenix);
      moteIF.registerListener(new BombillaErrorMsg(), assembler);
      injector = new ScriptInjector(moteIF, config.getVirusMap());
	
      /* Generate all of the code metadata */
      byte options;
      byte context;
      int version;

      options = 0;
      context = config.getCapsuleMap().nameToCode(contextName.toUpperCase());
      version = Integer.parseInt(versionString);

      /* Inject! */
      injector.inject(code, context, options, version);
    }
    catch (Exception ex) {
      ex.printStackTrace();
    }
  }
}
